@@ -1,7 +1,7 @@
 # -*- encoding: utf-8 -*-
 # vim: expandtab:tabstop=4:shiftwidth=4:softtabstop=4:autoindent
 
-"""nmevent v0.3.1 - C#-like implementation of the Observer pattern
+"""nmevent v0.3.2 - C#-like implementation of the Observer pattern
 
 This is a Python module :mod:`nmevent`, simple C#-like implementation of
 the Observer pattern (http://en.wikipedia.org/wiki/Observer_pattern).
@@ -119,21 +119,17 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 Changes
 =======
 
-v0.1
-  Initial release.
+v0.3.2
+  Added fairly useful functionality to the :func:`adapt` function: it
+  can now connect to "nested events". See the functions's documentation
+  for more information. It can also disconnect the handlers as well.
 
-v0.1.1
-  No changes in source code. Improved documentation and unit tests.
-
-v0.2
-  Rewritten most of the code. The :class:`Event` class now works as a 
-  descriptor, which eliminated the need for a separate :class:`EventSlot`
-  class and simplified usage. Added :class:`CallbackStore` to abstract
-  the callback storage. 
-
-v0.2.1
-  Rewritten some unit tests and added new ones. Improved documentation
-  a little bit.
+v0.3.1
+  Added docstring tests and fixed all the docstrings so that they
+  would pass. As a result, another problem was found with the event
+  binding. That problem was fixed by adding the :meth:`InstanceEvent.bind`
+  method to be used mainly by the :class:`Property` class when raising
+  the "changed" events.
 
 v0.3
   Fixed a major bug, which caused an unbound event not to be actually
@@ -143,12 +139,21 @@ v0.3
   decorates a class with "private" attributes for each property and
   automatic getters and setters where either one of them is missing.
 
-v0.3.1
-  Added docstring tests and fixed all the docstrings so that they
-  would pass. As a result, another problem was found with the event
-  binding. That problem was fixed by adding the :meth:`InstanceEvent.bind`
-  method to be used mainly by the :class:`Property` class when raising
-  the "changed" events.
+v0.2.1
+  Rewritten some unit tests and added new ones. Improved documentation
+  a little bit.
+
+v0.2
+  Rewritten most of the code. The :class:`Event` class now works as a 
+  descriptor, which eliminated the need for a separate :class:`EventSlot`
+  class and simplified usage. Added :class:`CallbackStore` to abstract
+  the callback storage. 
+
+v0.1.1
+  No changes in source code. Improved documentation and unit tests.
+
+v0.1
+  Initial release.
 """
 
 __version__ = __doc__.splitlines()[0].split(' ')[1][1:]
@@ -157,33 +162,82 @@ __all__    = [
     'nmproperty',
     'with_events',
     'with_properties',
+    'adapt',
     'Event',
 ]
 
 import __builtin__
+import weakref
 
 EVENTS_ATTRIBUTE = '__nmevents__'
 
+class WeakRefCallback(object):
+    @property
+    def is_alive(self):
+        return bool(self.callback and self.callback())
+    
+	def __init__(self, callback):
+		if hasattr(callback, "im_func"):
+			self.callback = weakref.ref(callback.im_self)
+			self.method = callback.im_func
+		else:
+			self.callback = weakref.ref(callback)
+			self.method = None
+	
+	def __repr__(self):
+		return "<WeakRefCallback(callback=%r, method=%r)>" % (self.callback, self.method)
+	
+	def __call__(self, *args, **keywords):
+		callback = self.callback and self.callback()
+		if not callback:
+			self.callback = None
+			return
+		if not self.method:
+			return callback(*args, **keywords)
+		return self.method.__get__(callback, type(callback))(*args, **keywords)
+    
+    def __hash__(self):
+        return hash((self.callback, self.method))
+    
+    def __eq__(self, other):
+        if not isinstance(other, WeakRefCallback):
+            return False
+        return hash(self) == hash(other)
+ 
 class CallbackStore(object):
     """Collection of callbacks."""
-
+    
     def __init__(self):
         """Constructor."""
         self.callbacks = set()
-
+    
     def __iter__(self):
         """Returns the collection's iterator object."""
         return iter(self.callbacks)
+    
+    def __iadd__(self, callback):
+        return self.add(callback)
 
+    def __isub__(self, callback):
+        return self.remove(callback)
+
+    def __contains__(self, callback):
+        return self.contains(callback)
+
+    def __len__(self):
+        return self.count()
+
+    def __call__(self, *args, **keywords):
+        return self.call(*args, **keywords)
+    
     def add(self, callback):
         """Adds a callback to callection.
-
+        
         :param callback: callable object to be added
         """
         self.callbacks.add(callback)
         return self
-    __iadd__ = add
-
+    
     def remove(self, callback):
         """Removes a callback from the collection.
 
@@ -191,30 +245,52 @@ class CallbackStore(object):
         """
         self.callbacks.remove(callback)
         return self
-    __isub__ = remove
-
+    
     def contains(self, callback):
         """Returns ``True`` is ``callback`` is in the collection.
 
         :param callback: callback to check for
         """
         return callback in self.callbacks
-    __contains__ = contains
-
+    
     def count(self):
         """Returns the number of callbacks in the collection."""
         return len(self.callbacks)
-    __len__ = count
-
+    
     def clear(self):
         """Removes all callbacks from collection."""
         self.callbacks = set()
-
+    
     def call(self, *args, **keywords):
         """Calls all callbacks with the given arguments."""
         for callback in self.callbacks:
             callback(*args, **keywords)
-    __call__ = call
+
+class WeakRefCallbackStore(CallbackStore):
+    def normalize(self, callback):
+        if isinstance(callback, WeakRefCallback):
+            return callback
+        return WeakRefCallback(callback)
+    
+    def __init__(self):
+        self.callbacks = {}
+    
+    def add(self, callback):
+        super(WeakRefCallbackStore, self).add(self.normalize(callback))
+
+    def remove(self, callback):
+        super(WeakRefCallbackStore, self).remove(self.normalize(callback))
+
+    def contains(self, callback):
+        super(WeakRefCallbackStore, self).contains(self.normalize(callback))
+
+    def call(self, *args, **keywords):
+        all_alive = True
+        for callback in self.callbacks:
+            callback(*args, **keywords)
+            all_alive = all_alive and callback.is_alive
+        if not all_alive:
+            self.callbacks = set([c for c in self.callbacks if c.is_alive])
 
 class Event(object):
     """Subject in the observer pattern.
@@ -784,3 +860,76 @@ def decorated(clss):
     """
     return with_events(with_properties(clss))
 
+def discover_handlers(observer, subject, prefix):
+    """Discovers event handlers for the subject's events in the observer.
+    
+    This function is not meant to be used directly. Use the :func:`adapt()`
+    function instead.
+    """
+    for attr in dir(observer):
+        if not attr.startswith(prefix):
+            continue
+        parts = attr[len(prefix):].split("__")
+        event = subject
+        while len(parts) > 0:
+            event = getattr(event, parts[0], None)
+            parts.pop(0)
+        if isinstance(event, InstanceEvent) or isinstance(event, Event):
+            yield (event, getattr(observer, attr))
+
+def adapt(observer, observable, prefix = "on_", disconnect = False):
+    """Connects observer's handlers to subject's events by their names.
+    
+    :param observer: object containing methods to connect as handler to the events
+    :param observable: object containing the events to connect to
+    :param prefix: prefix the names of handler methods start with
+    :param disconnect: when this is `True`, the function disconnects the handlers rather than connecting them
+    :returns: list of tuples of events and respective handlers
+    
+    Handler methods are recognized by their name. Their name must start with the
+    prefix and the continue with the name of the event.
+    
+    This function can also connect handlers to the events of subject's attributes,
+    even nested ones. To connect a handler to a "nested event", separate the
+    attribute names and the event name by double underscore ("__"). For instance,
+    to handle the "z" event of attribute "y" of attribute "x", name your handler
+    "on_x__y__z" (assuming you use the default prefix "on_").
+    
+    >>> class Observer(object):
+    ...    def on_x_happened(self, *senders, **keywords):
+    ...       print "x happened"
+    ... 
+    ...    def on_y_happened(self, *senders, **keywords):
+    ...       print "y happened"
+    ... 
+    ...    def on_attr__x_happened(self, *senders, **keywords):
+    ...       print "attr.x happened"
+    ...
+    >>> class Observable(object):
+    ...    x_happened = nmevent.Event()
+    ...    y_happened = nmevent.Event()
+    ...
+    >>> observer = Observer()
+    >>> observable = Observable()
+    >>> observable.attr = Observable()
+    >>> nmevent.adapt(observer, observable) #doctest: +ELLIPSIS
+    [...]
+    >>> observable.x_happened()
+    x happened
+    >>> observable.y_happened()
+    y happened
+    >>> observable.attr.x_happened()
+    attr.x happened
+    
+    """
+    handlers = list(discover_handlers(observer, observable, prefix))
+    if disconnect:
+        def fn(event, handler):
+            event -= handler
+    else:
+        def fn(event, handler):
+            event += handler
+    for event, handler in handlers:
+        fn(event, handler)
+    return handlers
+
