@@ -9,12 +9,28 @@ sys.path.insert(1, sys.path[0] + '/../nmevent')
 import nmevent
 
 suite = unittest.TestSuite()
+suite.addTests(doctest.DocFileSuite('../doc/index.rst', globs = {'nmevent': nmevent}))
 suite.addTests(doctest.DocTestSuite(nmevent, {'nmevent': nmevent}))
 
 def case(clss):
 	suite.addTests(unittest.defaultTestLoader.loadTestsFromTestCase(clss))
 	return clss
 
+def create_class(events = [], methods = [], attrs = {}):
+	class Clss(object):
+		def __init__(self):
+			self.counter = 0
+			for name, clss in attrs.items():
+				setattr(self, name, clss())
+	for event in events:
+		setattr(Clss, event, nmevent.Event())
+	for method in methods:
+		def meth(self, *args, **keywords):
+			self.counter += 1
+		meth.__name__ = method
+		setattr(Clss, method, meth)
+	return Clss
+
 def function_observer_a(sender, *args, **keywords):
 	pass
 
@@ -48,11 +64,48 @@ class CallableObserver(object):
 	def __init__(self):
 		self.event_caught = False
 		self.event_count = 0
-
+	
 	def __call__(self, *args, **keywords):
 		self.event_caught = True
 		self.event_count += 1
 
+@case
+class WeakRefCallbackTest(unittest.TestCase):
+	def setUp(self):
+		class Clss(object):
+			def __init__(self):
+				self.foo_counter = 0
+				self.bar_counter = 0
+			def foo(self):
+				self.foo_counter += 1
+			def bar(self):
+				self.bar_counter += 1
+		self.Clss = Clss
+		self.obj1 = self.Clss()
+		self.obj2 = self.Clss()
+		self.r1_foo = nmevent.WeakRefCallback(self.obj1.foo)
+		self.r1_bar = nmevent.WeakRefCallback(self.obj1.bar)
+		self.r2_foo = nmevent.WeakRefCallback(self.obj2.foo)
+		self.r2_bar = nmevent.WeakRefCallback(self.obj2.bar)
+	
+	def test_method(self):
+		self.assertEqual(self.obj1.foo_counter, 0)
+		self.r1_foo()
+		self.assertEqual(self.obj1.foo_counter, 1)
+		del self.obj1
+		self.r1_foo()
+	
+	def test_is_alive(self):
+		self.assertTrue(self.r1_foo.is_alive)
+		del self.obj1
+		self.assertFalse(self.r1_foo.is_alive)
+	
+	def test_hash(self):
+		self.assertEqual(hash(self.r1_foo),
+			hash(nmevent.WeakRefCallback(self.obj1.foo)))
+		self.assertNotEqual(hash(self.r1_foo), hash(self.r1_bar))
+		self.assertNotEqual(hash(self.r1_foo), hash(self.r2_foo))
+
 @case
 class CallbackStoreTest(unittest.TestCase):
 	def test_interface(self):
@@ -151,22 +204,22 @@ class CallbackStoreTest(unittest.TestCase):
 
 		self.assertEqual(len(store), 100)
 		self.assertEqual(store.count(), 100)
-
+		
 		store.clear()
-
+		
 		self.assertEqual(len(store), 0)
 		self.assertEqual(store.count(), 0)
-
+		
 		observer = Observer()
-
+		
 		store += observer.handler
-
+		
 		self.assertNotEqual(len(store), 0)
 		self.assertNotEqual(store.count(), 0)
 		self.assertTrue(observer.handler in store)
-
+		
 		store.clear()
-
+		
 		self.assertEqual(len(store), 0)
 		self.assertEqual(store.count(), 0)
 		self.assertFalse(observer.handler in store)
@@ -476,12 +529,113 @@ class WithPropertiesTest(unittest.TestCase):
 		a.bar = object()
 		self.assertTrue(a.bar is a._bar)
 		self.assertTrue(a.foo is a._foo)
+	
+	def test_with_events(self):
+		@nmevent.with_events
+		@nmevent.with_properties
+		class A(object):
+			foo = nmevent.Property()
+			bar = nmevent.Property()
+		
+		a = A()
+		b = A()
+		a.foo = object()
+		self.assertTrue(a.foo is not a.bar)
+		self.assertTrue(a.foo is not b.foo)
+		
+		observer_a = Observer()
+		observer_b = Observer()
+		a.foo_changed += observer_a.handler
+		a.bar_changed += observer_b.handler
+		a.foo = 1
+		a.foo = 2
+		a.foo = 3
+		self.assertEqual(observer_a.event_count, 3)
+		self.assertEqual(observer_b.event_count, 0)
 
-def do_test():
+@case
+class DiscoverHandlersTest(unittest.TestCase):
+	def test_simple(self):
+		subject = create_class(['event_a', 'event_b', ])()
+		observer = create_class([], ['on_event_a', 'on_event_b', ])()
+		
+		connections = list(nmevent.discover_handlers(observer, subject, "on_"))
+		
+		self.assertEqual(len(connections), 2)
+		self.assertTrue(isinstance(connections[0][0], nmevent.InstanceEvent))
+		self.assertEqual(connections[0][1], observer.on_event_a)
+		self.assertTrue(isinstance(connections[1][0], nmevent.InstanceEvent))
+		self.assertEqual(connections[1][1], observer.on_event_b)
+	
+	def test_nested(self):
+		subject = create_class(['some_event'], [], {
+			'attr1': create_class([], [], {
+				'attr2': create_class(['some_event', ])
+			}),
+		})()
+		observer = create_class([], ['on_attr1__attr2__some_event', ])()
+		
+		connections = list(nmevent.discover_handlers(observer, subject, "on_"))
+		
+		self.assertEqual(len(connections), 1)
+		self.assertEqual(connections[0][1], observer.on_attr1__attr2__some_event)
+
+@case
+class AdaptTest(unittest.TestCase):
+	def test_simple(self):
+		observer = create_class([], ['on_some_event', ])()
+		subject = create_class(['some_event', ])()
+		nmevent.adapt(observer, subject, "on_")
+		self.assertEqual(observer.counter, 0)
+		subject.some_event()
+		self.assertEqual(observer.counter, 1)
+		subject.some_event()
+		self.assertEqual(observer.counter, 2)
+	
+	def test_nested(self):
+		subject = create_class(['some_event'], [], {
+			'attr1': create_class([], [], {
+				'attr2': create_class(['some_event', ])
+			}),
+		})()
+		observer = create_class([], ['on_attr1__attr2__some_event', ])()
+		nmevent.adapt(observer, subject, "on_")
+		
+		self.assertEqual(observer.counter, 0)
+		subject.attr1.attr2.some_event()
+		subject.some_event()
+		self.assertEqual(observer.counter, 1)
+	
+	def test_disconnect(self):
+		subject = create_class(['some_event', 'other_event', ], [], {
+			'attr1': create_class([], [], {
+				'attr2': create_class(['some_event', ])
+			}),
+		})()
+		observer = create_class([], [
+			'on_some_event',
+			'on_other_event',
+			'on_attr1__attr2__some_event', ])()
+		nmevent.adapt(observer, subject, "on_")
+		self.assertTrue(
+			observer.on_attr1__attr2__some_event in subject.attr1.attr2.some_event)
+		self.assertTrue(
+			observer.on_some_event in subject.some_event)
+		self.assertTrue(
+			observer.on_other_event in subject.other_event)
+		nmevent.adapt(observer, subject, "on_", True)
+		self.assertFalse(
+			observer.on_attr1__attr2__some_event in subject.attr1.attr2.some_event)
+		self.assertFalse(
+			observer.on_some_event in subject.some_event)
+		self.assertFalse(
+			observer.on_other_event in subject.other_event)
+
+def run():
 	runner = unittest.TextTestRunner()
 	runner.run(suite)
 	# unittest.main()
 
 if __name__ == "__main__":
-	do_test()
+	run()
 
